import { expect } from 'chai';
import * as _ from 'lodash';
import { TestData } from '../utils/testData';
import { type TResolver } from '../../src/Resolver';
import { JsonTree } from '../../src/JsonTree';
import { Resolver } from '../../src/Resolver';

function testPerf(times: number, test: (run: (runner: () => void) => void) => void) {
  const perf = [];
  for (let i = 0; i < times; i++) {
    test((runner) => {
      const nowStart = performance.now();
      runner();
      const nowEnd = performance.now();
      perf.push(nowEnd - nowStart);
    });
  }
  return perf;
}

describe('Performance', () => {
  describe('graph + createShadowMap + createMetadataMap', () => {
    function createModel() {
      let resolvers = TestData;

      const setResolvers = (f: (previous: any) => any) => {
        resolvers = f(resolvers);
      };

      const handleDelete = (resolvers: TResolver[]) => {
        setResolvers((res) => {
          const r = { ...res };
          for (const k in r) {
            const isDel = resolvers.some((val) => k.startsWith(val));
            if (isDel) {
              delete r[k];
            }
          }
          return r;
        });
      };

      const handleResolverEdit = (previous: TResolver, next: TResolver) => {
        if (previous === next) {
          return;
        }

        setResolvers((res) => {
          return _.mapKeys(res, (_value, originalResolver) => {
            const updatedResolver = Resolver.replaceResolver(previous, next, [originalResolver])[0];
            return updatedResolver != null
              ? updatedResolver
              : originalResolver;
          });
        });
      };

      const handleDuplicate = (resolvers: TResolver[]) => {
        setResolvers((res) => {
          const r = { ...res };

          return r;
        });
      };

      const handleCreate = (resolvers: TResolver[]) => {
        setResolvers((res) => {
          const r = { ...res };
          resolvers.forEach((resolver) => {
            r[resolver] = {
              value: '',
              type: 'text',
            };
          });
          return r;
        });
      };

      return {
        get resolvers() {
          return resolvers;
        },
        handleDelete,
        handleDuplicate,
        handleCreate,
        handleResolverEdit,
      };
    }

    it('is performant', () => {
      const result = testPerf(5, (measure) => {
        const model = createModel();

        function createGraph() {
          const graph = Resolver.graph(model.resolvers);
          const shadowedGraph = JsonTree.createShadowMap(graph);
          JsonTree.createMetadataMap(shadowedGraph, graph);
        }

        function run() {
          createGraph();
          [
            () => model.handleCreate(['lang.{lang=pp}']),
            () => model.handleResolverEdit('lang.{lang=pp}.about.content', 'lang.{lang=pp}.changed.content'),
            () => model.handleDelete(['lang.{lang=pp}.changed.content']),
            () => model.handleDuplicate(['lang.{lang=pp}.changed.content']),
            () => model.handleResolverEdit('lang.{lang=en}', 'lang.{pop=en}'),
            // reverse
            () => model.handleResolverEdit('lang.{pop=en}', 'lang.{lang=en}'),
            () => model.handleDelete(['lang.{lang=pp}.changed.content']),
            () => model.handleCreate(['lang.{lang=pp}.changed.content']),
            () => model.handleResolverEdit('lang.{lang=pp}.changed.content', 'lang.{lang=pp}.about.content'),
            () => model.handleDelete(['lang.{lang=pp}']),
          ].forEach((action) => {
            action();
            createGraph();
          });
        }

        measure(() => {
          for (let i = 0; i < 10; i++) {
            run();
          }
        });
      });

      expect(result[0]).to.be.below(1300);
      expect(result[result.length - 1]).to.be.below(1000);
    }).timeout(5000);
  });

  describe('Resolver.split', () => {
    it('performance test', () => {
      const keys = _.keys(TestData);
      const result = testPerf(5, (measure) => {
        measure(() => {
          for (let i = 0; i < 5000; i++) {
            for (const key in keys) {
              Resolver.split(key);
            }
          }
        });
      });
      expect(result[0]).to.be.below(600);
      expect(result[result.length - 1]).to.be.below(600);
    });
  });

  describe('Resolver.graph', () => {
    it('performance test', () => {
      const result = testPerf(5, (measure) => {
        measure(() => {
          for (let i = 0; i < 50; i++) {
            Resolver.graph(TestData);
          }
        });
      });
      expect(result[0]).to.be.below(250);
      expect(result[result.length - 1]).to.be.below(190);
    });
  });

  describe('Resolver.filter', () => {
    it('performance test', () => {
      const result = testPerf(5, (measure) => {
        measure(() => {
          const matches = [
            ['a.b.{x=y}', 'a.b.{x=*}'],
            ['a.b.c.d', 'a.**.*'],
            ['a.b.c.d.e', 'a.**.e'],
          ];

          for (let i = 0; i < 500000; i++) {
            for (const [a, b] of matches) {
              Resolver.filter(a, b);
            }
          }
        });
      });

      for (const item of result) {
        expect(item).to.be.below(100)
      }
    });
  });
});
